! 
!  OpenMP lecture exercises
!  Copyright (C) 2011 by Christian Terboven <terboven@itc.rwth-aachen.de>
!            (C) 2015 by Paul       Kapinos  <kapinos@itc.rwth-aachen.de>
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
! 
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
! 
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
! 
! 
MODULE Tree
!$ USE OMP_LIB
use, intrinsic :: ISO_C_BINDING, only: C_INT
IMPLICIT NONE

interface 
  integer (C_INT) function rand()  BIND(C, name='rand')  
        use, intrinsic :: ISO_C_BINDING  
        implicit none 
  end function rand
end interface 

!  binary tree node
TYPE, PUBLIC :: SNode
  PRIVATE
  INTEGER :: value 							! node value
  TYPE(SNode), POINTER :: pleft  => NULL() 	! next left node
  TYPE(SNode), POINTER :: pright => NULL() 	! next right node
END TYPE SNode

CONTAINS


!  generate evenly balanced binary tree
RECURSIVE SUBROUTINE generate_tree(tree, depth, lev)
  TYPE(SNode), POINTER          :: tree
  INTEGER, INTENT(IN)           :: depth
  INTEGER, INTENT(IN), OPTIONAL :: lev
  INTEGER                       :: level
  IF (.NOT. PRESENT(lev) ) THEN
    level = 1
  ELSE
    level = lev
  END IF
  !
  IF (.NOT. ASSOCIATED(tree))    ALLOCATE (tree)
  IF (level < depth) THEN
    ALLOCATE (tree%pleft)
    ALLOCATE (tree%pright)
    CALL generate_tree(tree%pleft,  depth, level + 1);
    CALL generate_tree(tree%pright, depth, level + 1);
  END IF
END SUBROUTINE generate_tree


!  simple naming sheme
RECURSIVE SUBROUTINE name_tree(tree, depth, v)
  TYPE(SNode), POINTER          :: tree
  INTEGER, INTENT(IN)           :: depth, v
  IF ( .NOT. ASSOCIATED(tree%pleft) .AND. .NOT. ASSOCIATED(tree%pright) ) THEN
    !  assuming a complete binary tree, this is a leaf
    tree%value = MODULO( int(rand()), 8000000) + 1;
  ELSE
    !  first left branch
    CALL name_tree(tree%pleft, depth, v);
    !  then current node
    tree%value = MODULO( int(rand()), 8000000) + 1; 
    !  and finally the right branch
    CALL name_tree(tree%pright, depth, v);
  END IF
END SUBROUTINE name_tree


SUBROUTINE generate_and_name_tree(tree,  depth)
  TYPE(SNode), POINTER          :: tree
  INTEGER, INTENT(IN)           :: depth
  INTEGER                       :: value
  CALL generate_tree(tree, depth);
  value = 1;
  CALL name_tree(tree, depth, value);
END SUBROUTINE generate_and_name_tree


RECURSIVE SUBROUTINE DFS(tree,  v, found)
  TYPE(SNode), POINTER          :: tree
  INTEGER, INTENT(IN)           :: v
  LOGICAl, INTENT(INOUT)        :: found
  IF(tree%value == v)     found = .true.;
  IF( ASSOCIATED(tree%pright) ) CALL DFS(tree%pright,v,found);
  IF( ASSOCIATED(tree%pleft)  ) CALL DFS(tree%pleft, v,found);
END SUBROUTINE DFS


RECURSIVE SUBROUTINE parallelDFS( tree,  v, found)
  TYPE(SNode), POINTER          :: tree
  INTEGER, INTENT(IN)           :: v
  LOGICAl, INTENT(INOUT)        :: found
  !
  IF( tree%value == v )             found = .true.;
  IF( ASSOCIATED(tree%pright) ) CALL DFS(tree%pright,v,found);
  IF( ASSOCIATED(tree%pleft) )  CALL DFS(tree%pleft, v,found);
END SUBROUTINE parallelDFS

END MODULE Tree


PROGRAM Treesearch
  !$ USE OMP_LIB
  USE Tree
  IMPLICIT NONE
  DOUBLE PRECISION :: t1,t2 = 0.0
  INTEGER :: depth = 27, height = 0, elem = 42    !!! depth = 27
  TYPE(SNode), POINTER :: BinTree => NULL() 
  LOGICAL :: found
  !
  WRITE (*,*)  "Generating tree..."
  !$ t1 = omp_get_wtime();
     CALL generate_and_name_tree(BinTree,depth)
  !$ t2 = omp_get_wtime()-t1;
  WRITE (*,*)  " done in " , t2 , "sec";
  !
  found = .false.
  WRITE (*,*)  "Searching serially in the tree for value ", elem
  !$ t1 = omp_get_wtime();
    CALL DFS(BinTree,elem,found);
  !$ t2 = omp_get_wtime()-t1;
  IF (found .EQV. .true.) THEN
    WRITE (*,*)  "Element " , elem , " was found!";
  ELSE
    WRITE (*,*)  "Element " , elem , "was not found!";
  END IF
  WRITE (*,*)  "Searching was done in " , t2 , "sec";
  !
  found = .false.;
  WRITE (*,*)  "Searching in parallel in the tree for value " , elem;
  !$ t1 = omp_get_wtime();
    CALL parallelDFS(BinTree,elem,found);
  !$ t2 = omp_get_wtime()-t1;
  IF ( found .EQV. .true. ) THEN
    WRITE (*,*)  "Element " , elem , " was found!";
  ELSE
    WRITE (*,*)  "Element " , elem , "was not found!";
  END IF
  WRITE (*,*)  "Search in parallel was done in " , t2 , "sec";

END PROGRAM Treesearch
